#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <regex.h>

#include "config.h"
#include "imapfilter.h"
#include "connect.h"
#include "log.h"
#include "imap.h"


extern account_t *accounts;
extern unsigned int options;

int sock;


/*
 * Connects to mailserver.
 */
int init_connection(account_t * ca)
{
    struct sockaddr_in name;
    struct hostent *hostinfo;

    memset((char *) &name, 0, sizeof(struct sockaddr_in));

    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	error("imapfilter: create socket; %s\n", strerror(errno));
	return 1;
    }

    if (!(hostinfo = gethostbyname(ca->server))) {
	error("imapfilter: get network host entry; %s\n", strerror(errno));
	close_connection();
	return 1;
    }

    name.sin_family = AF_INET;
    name.sin_port = htons(ca->port);
    name.sin_addr = *(struct in_addr *) hostinfo->h_addr;

    if (connect(sock, (struct sockaddr *) &name, sizeof(struct sockaddr))) {
	error("imapfilter: initiating connection; %s\n", strerror(errno));
	close_connection();
	return 1;
    }

    info("Connected to %s as %s.\n", hostinfo->h_name, ca->userid);

    return 0;
}


/*
 * Disconnects from mailserver.
 */
int close_connection(void)
{
    if (close(sock)) {
	error("imapfilter: closing socket; %s\n", strerror(errno));
	return 1;
    } else
	return 0;
}


/*
 * Reads data the server sent.
 */
int recv_response(char *buf)
{
    int flags;
    fd_set rfds;
    struct timeval tv;

    memset(buf, 0, RESPONSE_BUF);

    tv.tv_sec = 10;
    tv.tv_usec = 0;

    flags = fcntl(sock, F_GETFL, 0);
    fcntl(sock, F_SETFL, flags | O_NONBLOCK);

    FD_ZERO(&rfds);
    FD_SET(sock, &rfds);

    if ((select(sock + 1, &rfds, NULL, NULL, &tv) > 0)
	&& FD_ISSET(sock, &rfds)) {
	if (read(sock, buf, RESPONSE_BUF - 1) != -1) {
#ifdef DEBUG
	    printf("\n%s\n", buf);
#endif
	    if (options & OPT_DETAILS_VERBOSE)
		analyze_response(buf);

	    fcntl(sock, F_SETFL, flags);
	    return 0;
	}
    }

    fcntl(sock, F_SETFL, flags);

    return 1;
}


/*
 * Clears stream from any data server sent.
 */
int clear_stream(void)
{
    char buf[RESPONSE_BUF];

    return recv_response(buf);
}


/*
 * Processes the data that server sent due to IMAP STATUS client request.
 */
int status_response(void)
{
    int r, s;
    char buf[RESPONSE_BUF];
    char exist[STATUS_BUF], recent[STATUS_BUF], unseen[STATUS_BUF];
    regex_t creg;
    regmatch_t match[4];
    const char *reg =
	"\\* STATUS INBOX \\(MESSAGES ([[:digit:]]+) RECENT ([[:digit:]]+) UNSEEN ([[:digit:]]+)\\)";

    exist[0] = recent[0] = unseen[0] = 0;

    r = recv_response(buf);

    regcomp(&creg, reg, REG_EXTENDED);

    if (!regexec(&creg, buf, 4, match, 0)) {
	s = min(match[1].rm_eo - match[1].rm_so, STATUS_BUF - 1);
	strncpy(exist, buf + match[1].rm_so, s);
	exist[s] = 0;

	s = min(match[2].rm_eo - match[2].rm_so, STATUS_BUF - 1);
	strncpy(recent, buf + match[2].rm_so, s);
	recent[s] = 0;

	s = min(match[3].rm_eo - match[3].rm_so, STATUS_BUF - 1);
	strncpy(unseen, buf + match[3].rm_so, s);
	unseen[s] = 0;

	info("%s messages exist, %s recent, %s unseen.\n", exist, recent,
	     unseen);
    }

    regfree(&creg);

    return r;
}


/*
 * Processes the data that server sent due to IMAP SEARCH client request.
 */
int search_response(char *res)
{
    int r, s, len;
    char buf[RESPONSE_BUF];
    regex_t creg;
    regmatch_t match[2];
    const char *reg = "\\* SEARCH([[:digit:] ]*)";

    len = strlen(res);

    r = recv_response(buf);

    regcomp(&creg, reg, REG_EXTENDED);

    if (!regexec(&creg, buf, 2, match, 0)) {
	s = min(match[1].rm_eo - match[1].rm_so, SEARCH_BUF - 1);
	strncpy(res + len, buf + match[1].rm_so, s);
	res[s + len] = 0;
    }

    regfree(&creg);

    return r;
}


/*
 * Processes the data that server sent due to IMAP FETCH client request.
 */
int fetch_response(char *res)
{
    int r, i, s, len = 0;
    char buf[RESPONSE_BUF];
    regex_t creg[3];
    regmatch_t match[1];
    const char *reg[3] = {
	"Date: [[:print:]]*",
	"From: [[:print:]]*",
	"Subject: [[:print:]]*"
    };

    r = recv_response(buf);

    for (i = 0; i < 3; i++) {
	regcomp(&creg[i], reg[i], REG_EXTENDED);

	if (!regexec(&creg[i], buf, 1, match, 0)) {
	    s = min(match[0].rm_eo - match[0].rm_so,
		    HEADERS_MAX - len - 1);
	    strncpy(res + len, buf + match[0].rm_so, s);
	    res[s + len] = '\n';
	    res[s + len + 1] = 0;
	    len = strlen(res);
	}
    }

    for (i = 0; i < 3; i++)
	regfree(&creg[i]);

    return r;
}


/*
 * Proccesses data that server sent due to IMAP EXPUNGE client request.
 */
int expunge_response(void)
{
    int r, i = 0, e = 0;
    char buf[RESPONSE_BUF];
    regex_t creg;
    regmatch_t match[1];
    const char *reg = "\\* [[:digit:]]+ EXPUNGE";

    r = recv_response(buf);

    regcomp(&creg, reg, REG_EXTENDED);

    while (!regexec(&creg, buf + e, 1, match, 0)) {
	e += match[0].rm_eo;
	i++;
    }
    
    info("%d messages expunged.\n", i);

    regfree(&creg);

    return r;
}


/*
 * Checks if response of server to client's request was succesfully
 * delivered or there was some kind of error.
 */
void analyze_response(char *buf)
{
    int s;
    regex_t creg;
    regmatch_t match[2];
    const char *reg = "[[:xdigit:]]{4,4} ((OK|NO|BAD) [[:print:]]*)";
    char res[RESULTS_BUF];

    regcomp(&creg, reg, REG_EXTENDED);

    if (!regexec(&creg, buf, 2, match, 0)) {
	s = min(match[1].rm_eo - match[1].rm_so, RESULTS_BUF - 1);
	strncpy(res, buf + match[1].rm_so, s);
	res[s] = 0;
	printf("Server response: %s\n", res);
    }

    regfree(&creg);
}
