#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/stat.h>

#include "config.h"
#include "imapfilter.h"
#include "file.h"
#include "log.h"


extern char logfile[PATH_MAX];
extern unsigned int options;

account_t *accounts = NULL;	/* All accounts. */

filter_t *dfilters = NULL;	/* Filters of DENY type. */
filter_t *afilters = NULL;	/* Filters of ALLOW type. */

unsigned int dlimit = 0;
unsigned int alimit = 0;	/* DENY and ALLOW message size limits. */


/*
 * Finds path to configuration file, opens it and calls parse_config().
 */
int read_config(char *cfg)
{
    int r;
    FILE *fp;
    char *home = NULL;

    if (!cfg) {
	cfg = (char *) malloc(PATH_MAX * sizeof(char));


	home = getenv("HOME");

	snprintf(cfg, PATH_MAX, "%s/%s", home, ".imapfilterrc");
    }
#ifdef DEBUG
    printf("debug: configuration file: '%s'\n", cfg);
#endif
#ifdef PERMISSIONS
    check_perms(cfg);
#endif
    fp = fopen(cfg, "r");

    if (!fp) {
	error("imapfilter: opening config file %s; %s\n",
	      cfg, strerror(errno));
	return 1;
    }

    r = parse_config(fp);

    fclose(fp);

#ifdef DEBUG
    printf("debug: options: %0#10x\n", options);
#endif

    return r;
}


#ifdef PERMISSIONS
/*
 * Checks the permissions of the configuration file.
 */
int check_perms(char *cfg)
{
    struct stat fs;

    if (stat(cfg, &fs)) {
	error("imapfilter: getting file %s status; %s\n", cfg,
	      strerror(errno));
	return 1;
    }

    if (!S_ISREG(fs.st_mode)) {
	error("imapfilter: file %s not a regular file\n", cfg);
	return 1;
    }

    if ((fs.st_mode & 00777) != (S_IRUSR | S_IWUSR)) {
	fprintf(stderr,
		"imapfilter: warning: bad config file %s permissions\n"
		"imapfilter: warning: file's mode should be 600 not %o\n",
		cfg, fs.st_mode & 00777);
	return 1;
    }

    return 0;
}
#endif

/*
 * Parses configuration file.
 */
int parse_config(FILE * fp)
{
    int i;
    unsigned int row = 0;
    char line[LINE_MAX];
    regex_t creg[7];
    regmatch_t match[4];
    const char *reg[7] = {
	"^[[:blank:]]*(SERVER|PORT|USERNAME|PASSWORD)[[:blank:]]*=[[:blank:]]*([[:graph:]]*)[[:blank:]]*\n$",
	"^[[:blank:]]*(DENY|ALLOW)[[:blank:]]*=[[:blank:]]*(FROM|CC|BCC|SUBJECT|TO):? ([[:print:]]*)\n$",
	"^[[:blank:]]*(DENY|ALLOW)[[:blank:]]*=[[:blank:]]*([[:graph:]]+):? ([[:print:]]*)\n$",
	"^[[:blank:]]*(DENY_LIMIT|ALLOW_LIMIT)[[:blank:]]*=[[:blank:]]*([[:digit:]]*)[[:blank:]]*\n$",
	"^[[:blank:]]*(LOGFILE)[[:blank:]]*=[[:blank:]]*([[:graph:]]*)[[:blank:]]*\n$",
	"^[[:blank:]]*(TEST_MODE|SHOW_HEADERS|UNSEEN_ONLY)[[:blank:]]*=[[:blank:]]*(yes|no)[[:blank:]]*\n$",
	"^([[:blank:]]*\n|#.*\n)$"
    };

    for (i = 0; i < 7; i++)
	regcomp(&creg[i], reg[i], REG_EXTENDED);

    while (fgets(line, LINE_MAX - 1, fp)) {
	row++;
	if (!regexec(&creg[0], line, 3, match, 0)) {
	    if (set_account(line, match))
		return 1;
	    continue;
	} else if (!regexec(&creg[1], line, 4, match, 0)) {
	    if (standard_filter(line, match))
		return 1;
	    continue;
	} else if (!regexec(&creg[2], line, 4, match, 0)) {
	    if (custom_filter(line, match))
		return 1;
	    continue;
	} else if (!regexec(&creg[3], line, 3, match, 0)) {
	    set_limits(line, match);
	    continue;
	} else if (!regexec(&creg[4], line, 3, match, 0) ||
		   !regexec(&creg[5], line, 3, match, 0)) {
	    set_options(line, match);
	    continue;
	} else if (!regexec(&creg[6], line, 0, match, 0)) {
	    continue;
	} else {
	    error("imapfilter: parse error in config file at row %d\n",
		  row);
	    return 1;
	}
    }

    /* Fail if no accounts were defined. */
    if (!accounts) {
	fprintf(stderr,
		"imapfilter: no accounts defined in config file\n");
	return 1;
    }

    /* Fail if no filters were defined. */
    if (!dfilters && !afilters) {
	error("imapfilter: no filters defined in config file\n");
	return 1;
    }

    for (i = 0; i < 7; i++)
	regfree(&creg[i]);

    return 0;
}


/*
 * Allocates memory for accounts, builds linked list, and NULL terminates
 * account settings.
 */
account_t *prepare_account(void)
{
    static account_t *ca;	/* Current account. */

    if (!accounts) {
	accounts = (account_t *) malloc(sizeof(account_t));
	if (!accounts) {
	    error("imapfilter: allocating memory; %s\n", strerror(errno));
	    return NULL;
	}
	ca = accounts;
    } else {
	ca->next = (account_t *) malloc(sizeof(account_t));
	if (!ca->next) {
	    error("imapfilter: allocating memory; %s\n", strerror(errno));
	    return NULL;
	}
	ca = ca->next;
    }

    ca->next = NULL;

    ca->server[0] = 0;
    ca->port = 143;
    ca->userid[0] = 0;
    ca->passwd[0] = 0;

    return ca;
}


/*
 * An account setting was found from parse_config(), changes the
 * apropriate variable.
 */
int set_account(char *line, regmatch_t * match)
{
    int s;
    char p[6];
    static account_t *ca;

    if (!strncmp(line + match[1].rm_so, "SERVER", 6)) {
	if (!(ca = prepare_account()))
	    return 1;
	s = min((match[2].rm_eo - match[2].rm_so), SERVER_MAX - 1);
	strncpy(ca->server, line + match[2].rm_so, s);
	ca->server[s] = 0;
#ifdef DEBUG
	printf("debug: account setting SERVER: '%s'\n", ca->server);
#endif
    } else if (accounts) {
	if (!strncmp(line + match[1].rm_so, "PORT", 4)) {
	    s = min((match[2].rm_eo - match[2].rm_so), sizeof(p));
	    strncpy(p, line + match[2].rm_so, s);
	    p[s] = 0;
	    ca->port = strtoul(p, NULL, 0);
#ifdef DEBUG
	    printf("debug: account setting PORT: %d\n", ca->port);
#endif
	} else if (!strncmp(line + match[1].rm_so, "USERNAME", 8)) {
	    s = min((match[2].rm_eo - match[2].rm_so), USERID_MAX - 1);
	    strncpy(ca->userid, line + match[2].rm_so, s);
	    ca->userid[s] = 0;
#ifdef DEBUG
	    printf("debug: account setting USERNAME: '%s'\n", ca->userid);
#endif
	} else if (!strncmp(line + match[1].rm_so, "PASSWORD", 8)) {
	    s = min((match[2].rm_eo - match[2].rm_so), PASSWD_MAX - 1);
	    strncpy(ca->passwd, line + match[2].rm_so, s);
	    ca->passwd[s] = 0;
#ifdef DEBUG
	    printf("debug: account setting PASSWORD: '%s'\n", ca->passwd);
#endif
	}
    }

    return 0;
}


/*
 * A filter entry was found from parse_config() processes it and saves it.
 */
int set_filters(char *line, regmatch_t * match, int csm)
{
    int s;
    static filter_t *cdf, *caf;
    filter_t **cf;
    filter_t **fl;

    if (!(strncmp(line + match[1].rm_so, "DENY", 4))) {
#ifdef DEBUG
	printf("debug: filter entry DENY:");
#endif
	cf = &cdf;
	fl = &dfilters;
    } else {
#ifdef DEBUG
	printf("debug: filter entry ALLOW:");
#endif
	cf = &caf;
	fl = &afilters;
    }

    /*
     * Allocate memory for filter entry and build linked list.
     */

    if (!(*fl)) {
	*fl = (filter_t *) malloc(sizeof(filter_t));
	if (!(*fl)) {
	    error("imapfilter: allocating memory; %s\n", strerror(errno));
	    return 1;
	}
	*cf = *fl;
    } else {
	(*cf)->next = (filter_t *) malloc(sizeof(filter_t));
	if (!((*cf)->next)) {
	    error("imapfilter: allocating memory; %s\n", strerror(errno));
	    return 1;
	}
	*cf = (*cf)->next;
    }

    (*cf)->next = NULL;

    (*cf)->custom = csm;

    s = min((match[2].rm_eo - match[2].rm_so), FIELD_NAME_MAX - 1);
    strncpy((*cf)->name, line + match[2].rm_so, s);
    (*cf)->name[s] = 0;

    s = min((match[3].rm_eo - match[3].rm_so), FIELD_BODY_MAX - 1);
    strncpy((*cf)->body, line + match[3].rm_so, s);
    (*cf)->body[s] = 0;

#ifdef DEBUG
    printf(" '%s' '%s'\n", (*cf)->name, (*cf)->body);
#endif
    return 0;
}


/*
 * Sets the DENY and ALLOW limits from the values found in config file.
 */
void set_limits(char *line, regmatch_t * match)
{
    int s;
    char lim[10];

    s = min((match[2].rm_eo - match[2].rm_so), sizeof(lim));
    strncpy(lim, line + match[2].rm_so, s);
    lim[s] = 0;

    if (!strncmp(line + match[1].rm_so, "DENY_LIMIT", 10)) {
	dlimit = strtoul(lim, NULL, 0);
#ifdef DEBUG
	printf("debug: size filter DENY_LIMIT: %d\n", dlimit);
#endif
    } else {
	alimit = strtoul(lim, NULL, 0);
#ifdef DEBUG
	printf("debug: size filter ALLOW_LIMIT: %d\n", alimit);
#endif
    }
}


/*
 * Sets other options found in config file.
 */
void set_options(char *line, regmatch_t * match)
{
    int s;

    if (!logfile[0] && !strncmp(line + match[1].rm_so, "LOGFILE", 7)) {
	s = min((match[2].rm_eo - match[2].rm_so), PATH_MAX - 1);
	strncpy(logfile, line + match[2].rm_so, s);
	logfile[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "TEST_MODE", 9) &&
	       !(options & OPT_TEST_MODE_STICKY)) {
	if (!strncmp(line + match[2].rm_so, "yes", 3))
	    options |= OPT_TEST_MODE;
	else if (options & OPT_TEST_MODE)
	    options ^= OPT_TEST_MODE;
    } else if (!strncmp(line + match[1].rm_so, "SHOW_HEADERS", 12)) {
	if (!strncmp(line + match[2].rm_so, "yes", 3))
	    options |= OPT_SHOW_HEADERS;
	else if (options & OPT_UNSEEN_ONLY)
	    options ^= OPT_SHOW_HEADERS;
    } else if (!strncmp(line + match[1].rm_so, "UNSEEN_ONLY", 11)) {
	if (!strncmp(line + match[2].rm_so, "yes", 3))
	    options |= OPT_UNSEEN_ONLY;
	else if (options & OPT_UNSEEN_ONLY)
	    options ^= OPT_UNSEEN_ONLY;
    }
}
