#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "config.h"
#include "imapfilter.h"
#include "imap.h"
#include "connect.h"
#include "log.h"


extern int sock;
extern account_t *accounts;
extern filter_t *dfilters, *afilters;
extern unsigned int options;
extern unsigned int dlimit, alimit;

static int tag = 0xA000;	/* Every IMAP command is prefixed with a
				   unique [:alnum:] string. */


/*
 * Sends to server data; a command.
 */
int send_command(char *cmd)
{

#ifdef DEBUG
    printf("debug: sending command: %s", cmd);
#endif

    if (write(sock, cmd, strlen(cmd)) == -1) {
	error("imapfilter: error while sending command; %s",
	      strerror(errno));
	return 1;
    }

    return 0;
}


#ifdef DEBUG
/*
 * IMAP NOOP: does nothing always succeeds.
 */
int imap_noop(void)
{
    char cmd[SMALL_CMD];

    verbose("Client request:  NOOP\n");

    snprintf(cmd, SMALL_CMD, "%X NOOP\r\n", tag++);

    if (!send_command(cmd) && !clear_stream())
	return 0;
    else
	return 1;
}
#endif


/*
 * IMAP LOGOUT: informs server that client is done.
 */
int imap_logout(void)
{
    char cmd[SMALL_CMD];

    verbose("Client request:  LOGOUT\n");

    snprintf(cmd, SMALL_CMD, "%X LOGOUT\r\n", tag++);

    if (!send_command(cmd) && !clear_stream())
	return 0;
    else
	return 1;
}


/*
 * IMAP LOGIN: identifies client to server.
 */
int imap_login(account_t * ca)
{
    char cmd[MEDIUM_CMD];

    verbose("Client request:  LOGIN\n");

    snprintf(cmd, MEDIUM_CMD, "%X LOGIN \"%s\" \"%s\"\r\n",
	     tag++, ca->userid, ca->passwd);

    if (!send_command(cmd) && !clear_stream())
	return 0;
    else
	return 1;
}


/*
 * IMAP SELECT: selects a mailbox in which messages can be accessed.
 */
int imap_select(void)
{
    char cmd[SMALL_CMD];

    verbose("Client request:  SELECT\n");

    snprintf(cmd, SMALL_CMD, "%X SELECT INBOX\r\n", tag++);

    if (!send_command(cmd) && !clear_stream())
	return 0;
    else
	return 1;
}


/*
 * IMAP STATUS: requests status of the indicated mailbox.
 */
int imap_status(void)
{
    char cmd[MEDIUM_CMD];

    verbose("Client request: STATUS\n");

    snprintf(cmd, MEDIUM_CMD,
	     "%X STATUS INBOX (MESSAGES RECENT UNSEEN)\r\n", tag++);

    if (!send_command(cmd) && !status_response())
	return 0;
    else
	return 1;
}


/*
 * IMAP SEARCH: searches the mailbox for messages that match certain criteria.
 */
int imap_search(char *res)
{
    char cmd[BIG_CMD];

    verbose("Client request:  SEARCH\n");

    gen_search_filters(cmd);
    if (send_command(cmd) || search_response(res))
	return 1;

    verbose("Client request:  SEARCH\n");

    if (alimit && afilters) {
	gen_search_limits(cmd);
	if (send_command(cmd) || search_response(res))
	    return 1;
    }

    return 0;
}


/*
 * IMAP FETCH: retrieves data associated with a message.
 */
int imap_fetch(unsigned int msg, char *res)
{
    char cmd[MEDIUM_CMD];

    verbose("Client request:  FETCH\n");

    snprintf(cmd, MEDIUM_CMD,
	     "%X FETCH %d BODY[HEADER.FIELDS (\"DATE\" \"FROM\" \"SUBJECT\")]\r\n",
	     tag++, msg);

    if (!send_command(cmd) && !fetch_response(res))
	return 0;
    else
	return 1;
}


/*
 * IMAP STORE: alters data associated with a message.
 */
int imap_store(unsigned int msg)
{
    char cmd[MEDIUM_CMD];

    verbose("Client request:  STORE\n");

    snprintf(cmd, MEDIUM_CMD,
	     "%X STORE %d +FLAGS \\Deleted\r\n", tag++, msg);

    if (!send_command(cmd) && !clear_stream())
	return 0;
    else
	return 1;
}


/*
 * IMAP EXPUNGE: premanently removes any messages with the \Deleted flag set.
 */
int imap_expunge(void)
{
    char cmd[SMALL_CMD];

    verbose("Client request:  EXPUNGE\n");

    snprintf(cmd, SMALL_CMD, "%X EXPUNGE\r\n", tag++);

    if (!send_command(cmd) && !expunge_response())
	return 0;
    else
	return 1;
}


/*
 * Prepares according to the filters defined by the user the IMAP SEARCH
 * command that will be sent.  This command will search for DENY, ALLOW,
 * and DENY_LIMIT filters.
 */
void gen_search_filters(char *cmd)
{
    int len;

    snprintf(cmd, BIG_CMD, "%X SEARCH", tag++);

    gen_deny_filters(cmd);

    if (dlimit) {
	len = strlen(cmd);
	snprintf(cmd + len, BIG_CMD - len, " LARGER %d", dlimit);
    } else if (!dfilters) {	/* If no DENY filters were defined, then
				   deny all except the ALLOW filters. */
	len = strlen(cmd);
	snprintf(cmd + len, BIG_CMD - len, " ALL");
    }

    gen_allow_filters(cmd, "NOT");

    if (options & OPT_UNSEEN_ONLY) {
	len = strlen(cmd);
	snprintf(cmd + len, BIG_CMD - len, " UNSEEN");
    }

    len = strlen(cmd);
    snprintf(cmd + len, BIG_CMD - len, "\r\n");
}


/*
 * Prepares according to the filters defined by the user the IMAP SEARCH
 * command that will be sent.  This command will search for the ALLOW_LIMIT
 * filter.
 */
void gen_search_limits(char *cmd)
{
    int len;

    snprintf(cmd, BIG_CMD, "%X SEARCH", tag++);

    gen_allow_filters(cmd, "OR");

    len = strlen(cmd);
    snprintf(cmd + len, BIG_CMD - len, " LARGER %d", alimit);

    if (options & OPT_UNSEEN_ONLY) {
	len = strlen(cmd);
	snprintf(cmd + len, BIG_CMD - len, " UNSEEN");
    }

    len = strlen(cmd);
    snprintf(cmd + len, BIG_CMD - len, "\r\n");

}


/*
 * Adds to the IMAP SEARCH command all the DENY type filters.
 */
void gen_deny_filters(char *cmd)
{
    int len;
    filter_t *cdf;

    if (dfilters) {
	for (cdf = dfilters; cdf->next; cdf = cdf->next) {
	    len = strlen(cmd);
	    snprintf(cmd + len, BIG_CMD - len,
		     " OR %s%s \"%s\"",
		     custom_header(cdf->custom), cdf->name, cdf->body);
	}

	len = strlen(cmd);
	snprintf(cmd + len, BIG_CMD - len, " %s%s%s \"%s\"",
		 (dlimit ? "OR " : ""),
		 custom_header(cdf->custom), cdf->name, cdf->body);
    }
}


/*
 * Adds to IMAP SEARCH command all the ALLOW type filters.
 */
void gen_allow_filters(char *cmd, char *key)
{
    int len;
    filter_t *caf;

    if (afilters) {
	for (caf = afilters; caf->next; caf = caf->next) {
	    len = strlen(cmd);
	    snprintf(cmd + len, BIG_CMD - len,
		     " %s %s%s \"%s\"", key,
		     custom_header(caf->custom), caf->name, caf->body);
	}

	len = strlen(cmd);
	snprintf(cmd + len, BIG_CMD - len, " %s%s%s \"%s\"",
		 (!strncmp(key, "NOT", 3) ? "NOT " : ""),
		 custom_header(caf->custom), caf->name, caf->body);
    }
}


/*
 * Converts the string with the UID's of the messages to be deleted to
 * integers, sends the appropriate command and optionally prints some
 * headers of the "to be deleted" messages.
 */
void del_messages(char *msgs)
{
    unsigned long int m;
    char hdr[HEADERS_MAX];

    do {
	m = strtoul(msgs, &msgs, 0);

	if (options & OPT_SHOW_HEADERS) {
	    imap_fetch(m, hdr);

	    if (!(options & OPT_DETAILS_QUITE))
		printf("Deleting message:\n%s", hdr);

	    if (!(options & OPT_TEST_MODE))
		log_info("%s", hdr);
	}

	if (!(options & OPT_TEST_MODE))
	    imap_store(m);

    } while (*msgs);
}
